/**************************************************************************
 *
 *  This file is part of the Coin GUI binding libraries.
 *  Copyright (C) 2000 by Systems in Motion.  All rights reserved.
 *
 *  The libraries this file is part of is free software; you can
 *  redistribute them and/or modify them under the terms of the GNU
 *  Lesser General Public License version 2.1 as published by the
 *  Free Software Foundation.  See the file LICENSE.LGPL at the root
 *  directory of the distribution for all the details.
 *
 *  If you want to use the Coin GUI binding libraries for applications
 *  not compatible with the LGPL, contact SIM about acquiring a
 *  Professional Edition License.
 *
 *  Systems in Motion, Prof Brochs gate 6, N-7030 Trondheim, NORWAY
 *  http://www.sim.no/ support@sim.no Voice: +47 22114160 Fax: +47 22207097
 *
 **************************************************************************/

// @configure_input@

/*!
  \class SoWin SoWin.h Inventor/Win/SoWin.h
  \brief The SoWin class takes care of Win initialization and event dispatching.
  \ingroup misc

  This is the "application-wide" class with solely static methods
  handling initialization and event processing tasks. You must use
  this class in any application built on top of the SoWin
  library. Typical usage is as follows (complete application code):

  \code
#include <Inventor/Win/SoWin.h>
#include <Inventor/Win/viewers/SoWinExaminerViewer.h>
#include <Inventor/nodes/SoCube.h>

int
main(int argc, char **argv)
{
  // Initialize SoWin and Open Inventor libraries. This returns a main
  // window to use.
  HWND mainwin = SoWin::init(argv[0]);

   // Make a dead simple scene graph, only containing a single cube.
  SoCube * cube = new SoCube;

  // Use one of the convenient viewer classes.
  SoWinExaminerViewer * eviewer = new SoWinExaminerViewer(mainwin);
  eviewer->setSceneGraph(cube);
  eviewer->show();

  // Pop up the main window.
  SoWin::show(mainwin);
  // Loop until exit.
  SoWin::mainLoop();
  return 0;
}
  \endcode

  \sa SoWinComponent
*/

#include <stdlib.h>
#include <string.h>
#include <Inventor/Win/SoWin.h>
#include <Inventor/Win/SoAny.h>


// Default constructor and the destructor is private (SoWin class
// consists solely of static functions).
SoWin::SoWin(void) { }
SoWin::~SoWin() { }


// *************************************************************************

/*!
  \fn HWND SoWin::init(const char * appname, const char * classname)

  This function initializes the SoWin library.

  The returned HWND is a toplevel shell widget for the
  application, which can be used as a shell for the main component.

  This method is provided for easier porting / compatibility with the
  original SGI Inventor InventorXt library. It just adds dummy \a argc
  and \a argv arguments and calls the SoWin::init() method below.
*/
HWND
SoWin::internal_init(const char * appname, const char * classname)
{
  // If this is SoQt, Qt will keep a reference to the appname string
  // -- so make it guaranteed permanent.
  char * buf = NULL;
  if (appname != NULL) {
    buf = new char[strlen(appname) + 1]; // FIXME: minor memleak, free on app-exit. 20020117 mortene.
    (void)strcpy(buf, appname);
  }

  // Fake argc and argv setup, forward to real init routine.
  char * array[] = { buf, NULL };
  int argc = (buf != NULL) ? 1 : 0;
  return SoWin::init(argc, array, appname, classname);
}

// *************************************************************************

/*!
  This method will fill in the integers pointed to by the arguments
  with the corresponding part of the version release number of the
  SoWin library.

  A \c NULL pointer will make that part of the version number be ignored.

  This method is not part of the original InventorXt API from SGI.
*/
void
SoWin::getVersionInfo(int * major, int * minor, int * micro)
{
  if (major) { *major = SOWIN_MAJOR_VERSION; }
  if (minor) { *minor = SOWIN_MINOR_VERSION; }
  if (micro) { *micro = SOWIN_MICRO_VERSION; }
}

/*!
  This method returns a string containing the version id of the
  library.

  This method is not part of the original InventorXt API from SGI.
*/
const char *
SoWin::getVersionString(void)
{
  static const char version[] = SOWIN_VERSION;
  return version;
}

// *************************************************************************

/*!
  \typedef void SoWin::FatalErrorCB(const SbString errmsg, SoWin::FatalErrors errcode, void * userdata)

  An application function callback for handling fatal errors should be
  of this type.

  The first argument is an error message in English describing the
  details of the error. The second argument is an error code used so
  the application can identify specific conditions. The third argument
  is the userdata pointer passed in to SoWin::setFatalErrorHandler().
*/

/*!
  \enum SoWin::FatalErrors
  Numerical identifiers for classifying the different kinds of possible
  fatal errors.
*/

/*!
  \var SoWin::FatalErrors SoWin::NO_OPENGL_CANVAS

  Could not construct \e any valid OpenGL canvas. Something is very
  wrong on the client system.
*/
/*!
  \var SoWin::FatalErrors SoWin::INTERNAL_ASSERT

  An internal error condition that should never happen was
  detected. The most likely cause of this is programmering errors
  within the SoWin library itself.
*/
/*!
  \var SoWin::FatalErrors SoWin::UNSPECIFIED_ERROR

  Signifies that we were not able to specify in any greater detail the
  error condition that came up.
*/

/*!
  Set up a callback to invoke in the case of unexpected fatal error
  conditions within the SoWin library.

  Almost any error condition within the library is handled in a robust
  way through return values indicating errors for the offending calls,
  but there are a few cases that are impossible to handle without
  seriously crippling the functionality.

  (One example is if we fail to find \e any way of making a valid
  OpenGL canvas. This is an indication that something is seriously
  wrong on the end-user's system, and the SoWin library will not work
  properly.)

  In the case of a fatal error, it is expected that the given
  application callback function communicates the problem to the
  end-user and then either exits the application or at least refrains
  from using any part of the SoWin library.

  If no callback is explicitly set up by the application, the SoWin
  library will display an error message to the end-user and then exit
  the application.

  When setting a callback, this method returns a pointer to the
  previous callback function, or \c NULL if none.

  (This is an extension versus the original SGI InventorXt library
  API.)


  On a related note, be aware that the end-user will still be notified
  about non-fatal errors and warning messages through a dialog box. If
  you want to handle these yourself, or if you don't want your
  end-user to see any of these non-critical messages, you will also
  need to install an error handler callback function on the Coin
  library itself:

  \code
    SoDebugError::setHandlerCallback(myErrorHandlerCB, myCBUserData);
  \endcode

  (Please also see the documentation of
  SoDebugError::setHandlerCallback().)

 */
SoWin::FatalErrorCB *
SoWin::setFatalErrorHandler(SoWin::FatalErrorCB * cb, void * userdata)
{
  return SoAny::si()->setFatalErrorHandler(cb, userdata);
}

// *************************************************************************

/*!
  Returns \c TRUE if this binary version of the SoWin library was
  compiled with debug settings on.

  This method was not part of the original SGI InventorXt library, but
  is an extension specific to the Coin project.
*/
SbBool
SoWin::isDebugLibrary(void)
{
  return SOWIN_DEBUG != 0 ? TRUE : FALSE;
}

/*!
  Returns \c TRUE if this version of the SoWin library has an
  Application Binary Interface compatible with the given version.

  This method was not part of the original SGI InventorXt library, but
  is an extension specific to the Coin project.
*/
SbBool
SoWin::isCompatible(unsigned int major, unsigned int minor)
{
  if (major != SOWIN_MAJOR_VERSION) { return FALSE; }
  if (minor > SOWIN_MINOR_VERSION) { return FALSE; }
  return TRUE;
}

/*!
  Returns an indication on how the library was compiled: as a dynamic
  library, or as a static library.

  This method was not part of the original SGI InventorXt library, but
  is an extension specific to the Coin project.
*/
SoWinABIType
SoWin::getABIType(void)
{
  // FIXME: replace this ugly shit with a configure-check to test if
  // we're on a DLL-capable platform. 20020118 mortene.
#if defined(WIN32) || defined(_WIN32) || defined(__WIN32__) || defined(__NT__)
#ifdef SOWIN_MAKE_DLL
  return SOWIN_DLL_ABI;
#else
  return SOWIN_LIB_ABI;
#endif
#endif
  return SOWIN_UNKNOWN_ABI;
}

/*!
  Abort the application due to some kind of mismatch in the ABI
  settings / configuration. This should hopefully help application
  programmers avoid shooting themselves in the foot by controlling
  certain run-time parameters in the client application versus what is
  expected by the SoWin library.

  If you're an application programmer, it is very likely that you only
  need to consider this an internal library method.

  This method was not part of the original SGI InventorXt library, but
  is an extension specific to the Coin project.
*/
void
SoWin::abort(SoWinABIError error)
{
  switch (error) {
  case SOWIN_LINKTIME_MISMATCH:
    SoWin::createSimpleErrorDialog(NULL, "Fatal Error",
                                     "Detected linktime mismatch error.");
    break;

  case SOWIN_LINKSTYLE_MISMATCH:
    SoWin::createSimpleErrorDialog(NULL, "Fatal Error",
                                     "Detected linkstyle mismatch error (DLL vs. LIB).");
    break;

  case SOWIN_RUNTIME_MISMATCH:
    SoWin::createSimpleErrorDialog(NULL, "Fatal Error",
                                     "Detected runtime mismatch error (versioning and ABI compatibility).");
    break;

  default:
    // FIXME: shouldn't this rather be an assert? 20020118 mortene.
    SoWin::createSimpleErrorDialog(NULL, "Fatal Error",
                                     "Unknown error in SoWin :(");
    break;
  }

  // FIXME: call fatal error handler in SoAny. 20020118 mortene.
  exit(-1);
}

// *************************************************************************
