/* 
 *  SbTessContour.cxx
 *
 *
 *  Created by Laurent Garnier on Mon Feb 14 2005.
 *
 */ 

// this :
#include <HEPVis/SbTessContour.h>

#include <HEPVis/SbGL.h>

extern "C" {

#ifdef WIN32
#else
#define CALLBACK
#endif

typedef GLvoid(CALLBACK *Func)();

void CALLBACK beginCallback(GLenum aWhich,GLvoid * aThis)
{
  SbTessContour* This = (SbTessContour*)aThis;
  This->setBeginType(aWhich);
  This->resetVertex();
}
void CALLBACK errorCallback(GLenum aErrorCode,GLvoid * aThis)
{
  const GLubyte* estring = gluErrorString(aErrorCode);
  fprintf(stderr, "Tessellation Error: %s\n", estring);
  SbTessContour* This = (SbTessContour*)aThis;
  This->setError(true);
}

void CALLBACK endCallback()
{
}

void CALLBACK vertexCallback(GLvoid *vertex,GLvoid* aThis)
{
  SbTessContour* This = (SbTessContour*)aThis;
  This->addVertex((double*)vertex);
}


/*  combineCallback is used to create a new vertex when edges intersect.
    coordinate location is trivial to calculate, but weight[4] may be used
    to average color, normal, or texture coordinate data.
    In this program, color is weighted. */
void CALLBACK combineCallback(GLdouble coords[3],
                     void* vertex_data[4],
                     GLfloat weight[4],
                     GLdouble **dataOut,
                     GLvoid* aThis)
{
  SbTessContour* This = (SbTessContour*)aThis;
  double* vertex = new double[3];
  vertex[0] = coords[0];
  vertex[1] = coords[1];
  vertex[2] = coords[2];
  This->combineTmps().push_back(vertex);
  *dataOut = vertex;
}

}


SbTessContour::SbTessContour()
:fVertexNumber(0)
,fBeginType(GL_TRIANGLES)
,fError(false)
{
}

SbTessContour::~SbTessContour() { 
}

void SbTessContour::setError(bool aError) {
  fError = aError;
}

void SbTessContour::setBeginType(int aType) {
  fBeginType = aType;
}

void SbTessContour::resetVertex() {
  fVertexNumber = 0;
}
std::vector<double*>& SbTessContour::combineTmps(){
  return fCombineTmps;
}

void SbTessContour::addVertex(const double *vertex)
{
  // GL_TRIANGLE_STRIP
  // Draws a connected group of triangles. One triangle is defined for each
  // vertex presented after the first two vertices. For odd n, vertices n,
  // n+1, and n+2 define triangle n. For even n, vertices n+1, n, and n+2
  // define triangle n. N-2 triangles are drawn.
  if (fBeginType == GL_TRIANGLE_STRIP) {
    fTmp.pointC[0] = vertex[0];
    fTmp.pointC[1] = vertex[1];
    fTmp.pointC[2] = vertex[2];

    if (fVertexNumber >=2 ) {
      fTriangles.push_back(fTmp);
    }
    int rest = fVertexNumber % 2;
    if(rest==1) {
      fTmp.pointA[0] = vertex[0];
      fTmp.pointA[1] = vertex[1];
      fTmp.pointA[2] = vertex[2];
    }
    else {
      fTmp.pointB[0] = vertex[0];
      fTmp.pointB[1] = vertex[1];
      fTmp.pointB[2] = vertex[2];
    }
    fVertexNumber++;
  }

  // GL_TRIANGLE_FAN
  // Draws a connected group of triangles. One triangle is defined for each
  // vertex presented after the first two vertices. Vertices 1, n+1,
  // and n+2 define triangle n. N-2 triangles are drawn.
  else if (fBeginType == GL_TRIANGLE_FAN) {
    if (fVertexNumber == 0) {
      fTmp.pointA[0] = vertex[0];
      fTmp.pointA[1] = vertex[1];
      fTmp.pointA[2] = vertex[2];
    }
    else {
      fTmp.pointC[0] = vertex[0];
      fTmp.pointC[1] = vertex[1];
      fTmp.pointC[2] = vertex[2];

      if (fVertexNumber >=2 ) {
        fTriangles.push_back(fTmp);
      }
      fTmp.pointB[0] = vertex[0];
      fTmp.pointB[1] = vertex[1];
      fTmp.pointB[2] = vertex[2];
    }
    fVertexNumber++;
  }
  // GL_TRIANGLES
  // Treats each triplet of vertices as an independent triangle.
  // Vertices 3n-2, 3n-1, and 3n define triangle n. N/3 triangles are drawn.
  else if (fBeginType == GL_TRIANGLES) {
    int rest = fVertexNumber % 3;
    if(rest==2) {
      fTmp.pointC[0] = vertex[0];
      fTmp.pointC[1] = vertex[1];
      fTmp.pointC[2] = vertex[2];

      fTriangles.push_back(fTmp);
    }
    else if(rest==1) {
      fTmp.pointB[0] = vertex[0];
      fTmp.pointB[1] = vertex[1];
      fTmp.pointB[2] = vertex[2];
    }
    else if(rest==0) {
      fTmp.pointA[0] = vertex[0];
      fTmp.pointA[1] = vertex[1];
      fTmp.pointA[2] = vertex[2];
    }
    fVertexNumber++;
  }
  else {
    // do nothing and should never happend
  }
}

std::vector<SbTessTriangle> SbTessContour::getFilledArea(const std::vector<std::vector<SbVec3f> > aContour) {

  fTriangles.clear();
  fCombineTmps.clear();
  fError = false;

  unsigned int vecSize = 0;
  for(unsigned int a=0;a<aContour.size();a++) {
    vecSize += aContour[a].size();
  }
  typedef GLdouble point[3];
  point* tab = new point[vecSize];

  GLUtesselator* tobj = gluNewTess();

  gluTessCallback(tobj, (GLenum)GLU_TESS_VERTEX_DATA,(Func)vertexCallback);
  gluTessCallback(tobj, (GLenum)GLU_TESS_BEGIN_DATA,(Func)beginCallback);
  gluTessCallback(tobj, (GLenum)GLU_TESS_END,(Func)endCallback);
  gluTessCallback(tobj, (GLenum)GLU_TESS_ERROR_DATA,(Func)errorCallback);
  gluTessCallback(tobj, (GLenum)GLU_TESS_COMBINE_DATA,(Func)combineCallback);
  
  gluTessProperty(tobj, (GLenum)GLU_TESS_WINDING_RULE,GLU_TESS_WINDING_ODD);

  gluTessBeginPolygon(tobj, this);
  vecSize = 0;
  for(unsigned int a=0;a<aContour.size();a++) {
    gluTessBeginContour(tobj);
    for(unsigned int b=0;b<aContour[a].size()-1;b++) {
      tab[vecSize][0] = aContour[a][b][0];
      tab[vecSize][1] = aContour[a][b][1];
      tab[vecSize][2] = aContour[a][b][2];
      gluTessVertex(tobj, tab[vecSize], tab[vecSize]);
      vecSize++;
    }
    gluTessEndContour(tobj);
  }
  gluTessEndPolygon(tobj);
  gluDeleteTess(tobj);

  delete [] tab;  

  for(unsigned int index=0;index<fCombineTmps.size();index++) {
    delete [] fCombineTmps[index];
  }
  fCombineTmps.clear();

  if(fError) fTriangles.clear();

  return fTriangles;
}

